from pprint import pprint
import pytest
from ejudge import io
from iospec import parse_string, iotypes


@pytest.fixture
def iospec():
    return parse_string(
        'name: <foo>\n'
        'hello foo!\n'
        '\n'
        'name: <bar>\n'
        'hello bar!')


@pytest.fixture
def lang():
    return 'python'


@pytest.fixture
def src_ok():
    return (
        'name = input("name: ")\n'
        'print("hello %s!" % name)')


@pytest.fixture
def src_bad():
    return (
        'name = input("name: ")\n'
        'print(name)')


#
# Test simple io.run() interactions
#
def test_run_ok(src_ok, lang, timeout=None, sandbox=False):
    tree = io.run(src_ok, ['foo'], lang=lang, sandbox=sandbox, timeout=timeout)
    case = tree[0]
    tree.pprint()
    assert tree
    assert len(tree) == 1
    assert case[0] == 'name: '
    assert case[1] == 'foo'
    assert case[2] == 'hello foo!'


def test_run_ok_with_timeout(src_ok, lang):
    test_run_ok(src_ok, lang, timeout=1.0)


def test_run_ok_with_sandbox(src_ok, lang):
    test_run_ok(src_ok, lang, sandbox=True)


def test_run_ok_with_sandbox_and_timeout(src_ok, lang):
    test_run_ok(src_ok, lang, sandbox=True, timeout=1.0)


#
# Test io.grade() and check if the feedback is correct
#
def test_grade_correct(iospec, src_ok, lang):
    feedback = io.grade(src_ok, iospec, lang=lang, sandbox=False)
    assert isinstance(feedback.answer_key, iotypes.TestCase)
    assert isinstance(feedback.case, iotypes.TestCase)
    assert feedback.grade == 1
    assert feedback.message is None
    assert feedback.status == 'ok'


def test_grade_wrong(iospec, src_bad, lang):
    feedback = io.grade(src_bad, iospec, lang=lang, sandbox=False)
    assert feedback.grade == 0
    assert feedback.status == 'wrong-answer'
    assert feedback.title == 'Wrong Answer'


if __name__ == '__main__':
    pytest.main('test_io_grader_python.py -v')



"""
Import builtin functions for easy recovery when a script or a preparation
routine messes with Python's __builtins__.
"""
import builtins

# Save this for later
__original__ = vars(builtins)
_print = print
_input = input


def update(D=None, **kwds):
    """Update the buitins module with the given values"""

    D = dict(D or {})
    D.update(kwds)
    for k, v in D.items():
        setattr(builtins, k, v)


def restore():
    """Restores builtins to original state"""

    for k, v in vars(builtins).items():
        if k not in __original__:
            delattr(builtins, k)
    update(__original__)



import os
import stat
import abc
import sys
import tempfile
import contextlib
import subprocess
import importlib
from boxed.pinteract import Pinteract
from iospec.iotypes import AttrDict
from iospec import iotypes
from iospec.runners import IoObserver
from iospec.util import indent
from ejudge import builtins
from ejudge import util


__all__ = ['IntegratedLanguage', 'ScriptingLanguage', 'CompiledLanguage',
           'BuildError', 'manager_from_lang', 'lang_from_extension']
_print = print
_stdout = sys.stdout
lang_mapping = {
}


class BuildError(Exception):
    pass


class LanguageManager:
    """
    Support for new languages is done by subclassing this class.

    The LanguageManager automatically register new classes using meta
    information that is passed as class attributes. It requires that the
    following attributes are defined:

    name : str
        Short lowercase name for the language. (e.g.: python)
    description : str
        A short description of the language (appending version numbers and
        possibly compile flags). (e.g.: Python 3.x)
    extensions:
        A list of valid extensions associated to that language.
    """

    abstract = True
    supported_languages = {}
    registered_extensions = {}
    registered_language_extensions = {}
    buildargs = None
    shellargs = None
    _send_early_termination_error = False

    def __init__(self, source, is_sandboxed=True):
        self.source = source
        self.context = None
        self.is_sandboxed = is_sandboxed

    @classmethod
    def from_language(cls, lang, source):
        try:
            factory = cls.supported_languages[lang]
        except KeyError:
            raise ValueError('invalid language: %r' % lang)
        return factory(source)

    @abc.abstractmethod
    def build_context(self):
        """Returns the build context for the given manager.

        This function is only called if self.syntax_check() has passed.

        Failed builds can either return None or raise a BuildError."""

        raise NotImplementedError

    @abc.abstractmethod
    def exec(self, inputs, context):
        """Executes the source code in the given context and return the
        resulting test case object

        The context parameter is the result of calling the build function."""

        raise NotImplementedError

    @abc.abstractmethod
    def syntax_check(self):
        raise SyntaxError('invalid syntax')

    def build(self):
        """Builds and check the source code syntax. Also prepares the
        environment before execution.

        This function can be called multiple times and it will return a cached
        build context."""

        if self.context is None:
            try:
                self.syntax_check()
            except SyntaxError as ex:
                raise BuildError(ex)

            self.context = self.build_context()
            if self.context is None:
                data = indent(self.source)
                raise BuildError('cannot build source: \n%s' % data)
        return self.context

    def reset_context(self):
        """Reset/reconfigure the self.context attribute after each run.

        The default implementation does nothing."""

    def prepare_error(self, ex):
        """Prepares an exception receiving trying to execute
        manager.exec(ctx, inputs) before it is inserted into the error attribute
        of the returning ErrorTestCase."""

        return str(ex)

    def flush_io(self):
        """Flush all io buffers and return what has been collected so far.

        The resulting value is a list of plain input or output nodes."""

        return []

    def run(self, inputs, *, timeout=None, context=None):
        """Executes the program with the given inputs"""

        inputs = map(str, inputs)
        if context is None:
            context = self.build()

        try:
            func = self.exec
            result = remove_trailing_newline(
                    util.timeout(func, args=(inputs, context), timeout=timeout))

        except TimeoutError as ex:
            msg = 'Maximum execution time exceeded: %s sec' % timeout
            result = iotypes.ErrorTestCase.timeout(
                data=self.flush_io(),
                error_message=msg)

        # Reset context before returning
        self.reset_context()
        return result

    def modules(self):
        """Return a list of module imports that should be passed to the
        boxed.run() function in sandboxed mode."""

        return [self.__module__]

    @contextlib.contextmanager
    def keep_cwd(self):
        """A utility function that implements a context manager that restores
        the cwd at the beginning of execution when it exits the with block."""

        currpath = os.getcwd()
        yield
        os.chdir(currpath)


# noinspection PyAbstractClass
class IntegratedLanguage(LanguageManager):
    """Base class for all languages that expose hooks for printing and input
    directly to Python's interpreter itself.

    This runner is the most integrated with the ejudge system. Ideally all
    languages should be implemented as subclasses of IntegratedLanguage.
    This may not be feasible or practical for most programming languages,
    though."""

    abstract = True

    def __init__(self, source, globals=None, locals=None, builtins=None):
        super().__init__(source)
        self.observer = IoObserver()
        self.globals = dict(globals or {})
        self.locals = dict(locals or {})
        self.builtins = {
            'print': self.observer.print,
            'input': self.observer.input,
        }
        self.builtins.update(builtins or {})

    def build_context(self):
        return AttrDict(
            globals=self.globals.copy(),
            locals=self.locals.copy(),
        )

    def reset_context(self):
        self.context.globals = self.globals.copy()
        self.context.locals = self.locals.copy()

    def flush_io(self):
        return self.observer.flush()

    def run(self, inputs, **kwds):
        builtins.update(self.builtins)
        self.observer.flush()
        self.observer.extend_inputs(inputs)

        try:
            result = super().run(inputs, **kwds)
        finally:
            builtins.restore()

        return remove_trailing_newline(result)

    def syntax_check(self):
        raise NotImplementedError('no syntax check implemented for %s' %
                                  type(self).__name__)

    def print(self, *args, file=_stdout, **kwds):
        return _print(*args, file=file, **kwds)


class ExternalExecution(LanguageManager):
    """Features common to scripting and compiled languages or any language in
    which the runtime is outside the Python interpreter."""

    abstract = True

    @property
    def extension(self):
        return self.extensions[0]

    @property
    def shellargs(self):
        raise RuntimeError('shellargs must be overriden in the subclass')

    def exec(self, inputs, context):
        return self.exec_pinteract(inputs, context)

    def exec_pinteract(self, inputs, context):
        """Run script as a subprocess and gather results of execution.

        This is a generic implementation. Specific languages or runtimes are
        supported by fixing the shellargs argument of this function.
        This function uses the PInteract() object for communication with
        scripts.

        This function may change the current working path.
        """

        def append_non_empty_out():
            data = process.receive()
            if data:
                result.append(iotypes.Out(data))

        tmpdir = context.tempdir
        result = iotypes.IoTestCase()

        # Execute script in the tmpdir and than go back once execution has
        # finished
        olddir = os.getcwd()
        os.chdir(tmpdir)

        try:
            process = Pinteract(self.shellargs)

            # Fetch all In/Out strings
            append_non_empty_out()
            for inpt in inputs:
                try:
                    process.send(inpt)
                    result.append(iotypes.In(inpt))
                except RuntimeError as ex:
                    # Early termination: we still have to decide if an specific
                    # early termination error should exist.
                    #
                    # The default behavior is just to send a truncated
                    # IoTestCase
                    if process.is_dead():
                        if self._send_early_termination_error:
                            missing = [inpt]
                            missing.extend(inputs)
                            missing = '\n'.join('    ' + x for x in missing)
                            msg = ('process closed with consuming all inputs. '
                                   'List of unused inputs:\n')

                            return iotypes.ErrorTestCase.earlytermination(
                                list(result),
                                error_message=msg + missing)
                        else:
                            return iotypes.IoTestCase(list(result))

                    # This clause is just a safeguard. We don't expect to ever
                    # get here
                    return iotypes.ErrorTestCase.runtime(
                        list(result),
                        error_message='An internal error occurred while trying '
                                      'to interact with the script: %s.' % ex
                    )
                append_non_empty_out()
        finally:
            if not self.is_sandboxed:
                os.chdir(olddir)

        # Finish process
        error_ = process.finish()
        assert not any(error_), error_
        return result


# noinspection PyAbstractClass
class ScriptingLanguage(ExternalExecution):
    """Basic support for scripting language."""

    abstract = True

    def build_context(self):
        """Base buildfunc for source code that can be executed as scripts.

        Concrete implementations must provide a default extension for the script
        files and a syntax_check(src) function, that takes a string of code and
        raise BuildProblem if the syntax is invalid."""

        tmpdir = tempfile.mkdtemp()
        tmppath = os.path.join(tmpdir, 'main.' + self.extension)

        # Make tmp directory readable and writable by everyone
        if self.is_sandboxed:
            os.chmod(tmpdir, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE |
                             stat.S_IXOTH | stat.S_IROTH | stat.S_IWOTH |
                             stat.S_IXGRP | stat.S_IRGRP | stat.S_IWGRP)

        # Save source and make it readable by everyone
        with open(tmppath, 'w') as F:
            F.write(self.source)
        if self.is_sandboxed:
            os.chmod(tmppath, stat.S_IREAD | stat.S_IROTH | stat.S_IRGRP)

        return AttrDict(
            tempdir=tmpdir,
            shellargs=self.shellargs,
            messages='',
        )


class CompiledLanguage(ExternalExecution):
    abstract = True
    shellargs = ['./main.exe']

    def build_context(self):
        """Base build function for source code that must be compiled as part of the
        build process.

        Pass the list of shell arguments to run a command that takes a "main.ext"
        file and compiles it to "main.exe". Raises a BuildProblem if compilation
        return with a non-zero value."""

        # Save source file in temporary dir
        tmpdir = tempfile.mkdtemp()
        tmppath = os.path.join(tmpdir, 'main.' + self.extension)
        with open(tmppath, 'w') as F:
            F.write(self.source)

        # Compile and return
        currpath = os.getcwd()
        errmsgs = 'compilation is taking too long'
        try:
            os.chdir(tmpdir)
            errmsgs = subprocess.check_output(self.buildargs, timeout=10)

            # Make executable readable and executable by everyone
            if self.is_sandboxed:
                os.chmod('main.exe',
                         stat.S_IREAD | stat.S_IROTH | stat.S_IRGRP |
                         stat.S_IEXEC | stat.S_IXOTH | stat.S_IXGRP)

        except (subprocess.CalledProcessError, TimeoutError):
            raise BuildError(errmsgs)
        finally:
            # Cannot return to path in sandboxed mode since the nobody user
            # probably does not have read permissions on the old working
            # directory
            if not self.is_sandboxed:
                os.chdir(currpath)

        return AttrDict(
            tempdir=tmpdir,
            shellargs=self.shellargs,
            messages=errmsgs,
        )


def remove_trailing_newline(case):
    """Remove a trailing newline from the last output node."""

    if case is None:
        return None

    if case:
        last = case[-1]
        if isinstance(last, iotypes.Out) and last.endswith('\n'):
            last.data = str(last[:-1])
    return case


#
# API function
#
ext_map = {}
lang_map = {}


def manager_from_lang(lang, src, **kwds):
    """Return a manager instance for the given language string.

    It passes the code source and any other keyword arguments that might be
    necessary to the correct initialization of the manager."""

    manager = lang_map[lang]

    if isinstance(manager, str):
        mod_name, _, manager_name = manager.rpartition('.')
        mod = importlib.import_module(mod_name)
        manager = getattr(mod, manager_name)

    return manager(src, **kwds)


def lang_from_extension(ext):
    """Return a lang string from the given extension."""

    if ext.startswith('.'):
        ext = ext[1:]
    return ext_map[ext]


def register_extension(ext, lang, force=False):
    """Map an extension to a given lang."""

    if ext.startswith('.'):
        ext = ext[1:]
    if force and ext in ext_map and ext != ext_map[lang]:
        raise RuntimeError('extension %r is already registered to %r' %
                           ('.' + ext, lang))
    ext_map[ext] = lang


def register_lang(lang, cls=None, force=False, extensions=(), aliases=()):
    """Register a LanguageManager class `cls` tho the given language `lang`.

    This function accepts either the class itself or a named reference such
    as "ejudge.langs.lang_c.GccManager".

    If force=True, it overwrites any manager class already registered to a
    language."""

    if cls is None:
        def decorator(cls):
            register_lang(lang, cls, force=force)
            return cls

    if force and lang in lang_map and lang_map[lang] != cls:
        raise RuntimeError('%r is already registered to %r' % (lang, cls))

    for ext in extensions:
        register_extension(ext, lang, force=force)

    lang_map[lang] = cls
    for alias in aliases:
        lang_map[alias] = cls


#
# Explicitly register supported languages
#

# C family languages
register_lang('c', 'ejudge.langs.lang_c.GccManager',
              extensions=['.c'], aliases=['gcc'])
register_lang('tcc', 'ejudge.langs.lang_c.TccManager')
register_lang('clang', 'ejudge.langs.lang_c.ClangManager')
register_lang('c++', 'ejudge.langs.lang_c.GccCppManager',
              extensions=['.cpp'], aliases=['cpp', 'g++', 'gpp'])
register_lang('clang-c++', 'ejudge.langs.lang_c.ClangCppManager',
              aliases=['clang-cpp', 'clang++', 'clangpp'])

# Python family
register_lang('python', 'ejudge.langs.lang_python.PythonManager',
              extensions=['.py', 'py3'], aliases=['python3', 'py3', 'py'])
register_lang('python2', 'ejudge.langs.lang_python.Python2Manager',
              extensions=['.py2'], aliases=['python2', 'py2'])
register_lang('python-script', 'ejudge.langs.lang_python.PythonScriptManager')


# Exotic and special purpose languages
register_lang('pytuga', 'ejudge.langs.lang_pytuga.PytugaManager',
              extensions=['.pytg'], aliases=['pytg', 'pytugues', 'pytuguês'])


import pytest
from ejudge import io
from iospec import parse_string, iotypes
from ejudge.test.test_io_grader_python import (
    iospec,
    test_run_ok,
    test_run_ok_with_timeout,
    test_run_ok_with_sandbox,
    test_run_ok_with_sandbox_and_timeout,
)


@pytest.fixture
def lang():
    return 'gcc'


@pytest.fixture
def src_ok():
    return r"""
#include<stdio.h>

void main() {
    char buffer[100];
    printf("name: ");
    scanf("%s", buffer);
    printf("hello %s!\n", buffer);
}
"""


@pytest.fixture
def src_bad():
    return """
#include<stdio.h>

void main() {
    char buffer[100];
    puts("name: ");
    scanf("%s", buffer);
    printf("%s\n", buffer);
}
"""

if __name__ == '__main__':
    pytest.main('test_io_grader_c.py')



from ejudge.langs import CompiledLanguage


#
# C compilers
#
# We have manager classes for the two biggest compilers gcc and clang and also
# for tcc, the tiny C compiler, which is a small and fast compiler. In the
# future we might create an IntegratedManager based on clang or tcc or even on
# a C interpreter like ROOT's Cint.
#
class CLanguage(CompiledLanguage):
    abstract = True

    def syntax_check(self):
        pass


class TccManager(CLanguage):
    name = 'tcc'
    description = 'ANSI C (tcc)'
    extension = 'c'
    extensions = []
    buildargs = ['tcc', '-w', '-o', 'main.exe', 'main.c']


class GccManager(CLanguage):
    name = 'gcc'
    description = 'C99 (gcc)'
    extension = 'c'
    extensions = ['c']
    buildargs = ['gcc', '-lm', 'main.c', '-o', 'main.exe']


class ClangManager(CLanguage):
    name = 'clang'
    description = 'ANSI C (tcc)'
    extension = 'c'
    extensions = []
    buildargs = ['clang', '-lm', '-o', 'main.c', '-o', 'main.exe']


#
# C++ compilers
#
class CppLanguage(CompiledLanguage):
    abstract = True

    def syntax_check(self):
        pass


class GccCppManager(CppLanguage):
    name = 'g++'
    description = 'C++11 (gcc)'
    extension = 'cpp'
    extensions = ['cpp']
    buildargs = ['g++', '-lm', '-o', 'main.cpp', '-o', 'main.exe']


class ClangCppManager(CppLanguage):
    name = 'clang-c++'
    description = 'ANSI C (tcc)'
    extension = 'cpp'
    extensions = []
    buildargs = ['clang', '-lm', '-o', 'main.cpp', '-o', 'main.exe']



import pytest
from iospec import parse_string, iotypes
from ejudge import io
from ejudge.test.test_io_grader_python import (
    iospec,
    test_run_ok,
    test_run_ok_with_timeout,
    test_run_ok_with_sandbox,
    test_run_ok_with_sandbox_and_timeout,
    test_grade_correct,
    test_grade_wrong,
    src_ok,
    src_bad
)


@pytest.fixture
def lang():
    return 'python-script'


if __name__ == '__main__':
    pytest.main('test_io_grader_python_script.py')



import pytuga
from iospec import iotypes
from ejudge.langs import ScriptingLanguage, IntegratedLanguage


class PytugaManager(IntegratedLanguage, ScriptingLanguage):
    # abstract = True
    name = 'pytuga'
    description = 'Pytuguês'
    extensions = ['pytg']
    shellargs = ['pytuga', 'main.pytg']

    def syntax_check(self):
        return NotImplemented

    @property
    def _base(self):
        return IntegratedLanguage if self.is_sandboxed else ScriptingLanguage

    def build_context(self):
        return self._base.build_context(self)

    def exec_integrated(self, inputs, context):
        assert context is not None
        pytuga.exec(self.source, context.globals, context.locals,
                    forbidden=True)
        return iotypes.IoTestCase(self.flush_io())

    def exec(self, inputs, context):
        if self.is_sandboxed:
            return self.exec_integrated(inputs, context)
        else:
            return ScriptingLanguage.exec(self, inputs, context)

    def run(self, inputs, **kwds):
        return self._base.run(self, inputs, **kwds)

    def modules(self):
        return super().modules() + ['pytuga.lib.forbidden']



import pytest
from ejudge import io
from iospec import parse_string, iotypes


#
# Prevent regressions by registering specific interaction bugs
#
def test_less_inputs_than_expected(lang='python', sandbox=False):
    src = 'print(42)'
    iosrc = 'foo<42>\nbar<0>\n42'
    fb = io.grade(src, iosrc, lang=lang, sandbox=sandbox)
    fb.case.pprint()
    fb.answer_key.pprint()
    assert fb.grade == 0
    assert fb.status == 'wrong-answer'


def test_less_inputs_than_expected_sandbox():
    test_less_inputs_than_expected(sandbox=True)


def test_less_inputs_than_expected_pytuga():
    test_less_inputs_than_expected(lang='python3')


if __name__ == '__main__':
    pytest.main('test_examples.py')
